# -*- coding: utf-8 -*-

"""RabbitHole: Store messages from rabbitmq into a SQL database.

The way the message are stores is that each exchange name is mapped to a SQL
query that is executed when neded.

"""

__author__ = """Javier Collado"""
__email__ = 'javier@gigaspaces.com'
__version__ = '0.2.0'



# -*- coding: utf-8 -*-

"""Batcher: group messages in batches before writing them to the database.

The strategy to batch messages is:
    - store them in memory as they are received
    - send them to the database when either the size or the time limit is
    exceeded.

"""

import logging
import threading

from collections import defaultdict

import blinker

LOGGER = logging.getLogger(__name__)


class Batcher(object):

    """Group messages in batches before writing them to the database.

    A batch is queued for writing into the database when either the size or the
    time limit is exceeded.

    :param database: Database to use to insert message batches
    :type database: rabbithole.db.Database

    """

    DEFAULT_SIZE_LIMIT = 5
    DEFAULT_TIME_LIMIT = 15

    def __init__(self, size_limit, time_limit):
        """Initialize internal data structures."""
        self.size_limit = size_limit or self.DEFAULT_SIZE_LIMIT
        self.time_limit = time_limit or self.DEFAULT_TIME_LIMIT

        self.batches = defaultdict(list)
        self.locks = defaultdict(threading.Lock)
        self.timers = {}
        self.batch_ready = blinker.Signal()

    def message_received_cb(self, sender, exchange_name, payload):
        """Handle message received event.

        This callback is executed when message is received by the AMQP
        consumer.

        :param sender: The consumer who sent the message
        :type sender: rabbithole.consumer.Consumer
        :param exchange_name: Key used to determine which query to execute
        :type exchange_name: str
        :param payload: Rows to insert in the database
        :type payload: list(dict(str))

        """
        # Use a lock to make sure that callback execution doesn't interleave
        with self.locks[exchange_name]:
            batch = self.batches[exchange_name]
            batch.append(payload)
            LOGGER.debug(
                'Message added to %r batch (size: %d, capacity: %d)',
                exchange_name,
                len(batch),
                self.size_limit,
            )

            if len(batch) == 1:
                self.start_timer(exchange_name)
            elif len(batch) >= self.size_limit:
                LOGGER.debug(
                    'Size limit (%d) exceeded for %r',
                    self.size_limit,
                    exchange_name,
                )
                self.queue_batch(exchange_name)
                self.cancel_timer(exchange_name)

    def time_expired_cb(self, exchange_name):
        """Handle time expired event.

        This callback is executed in a timer thread when the time limit for a
        batch of messages has been exceeded.

        :param exchange_name: Exchange from which message batch was received
        :type exchange_name: str

        """
        # Use a lock to make sure that callback execution doesn't interleave
        with self.locks[exchange_name]:
            LOGGER.debug(
                'Time limit (%.2f) exceeded for %r',
                self.time_limit,
                exchange_name,
            )
            self.queue_batch(exchange_name)
            if exchange_name not in self.timers:
                LOGGER.warning('Timer not found for: %r', exchange_name)
                return
            del self.timers[exchange_name]

        thread = threading.current_thread()
        LOGGER.debug(
            'Timer thread finished: (%d, %s)',
            thread.ident,
            thread.name,
        )

    def queue_batch(self, exchange_name):
        """Queue batch for writing into database.

        A batch is queued either by the main thread when the size limit is
        exceeded or by a timer thread when the time limit is exceeded.

        :param exchange_name: Exchange from which message batch was received
        :type exchange_name: str

        """
        batch = self.batches[exchange_name]
        if not batch:
            LOGGER.warning('Nothing to queue for %r', exchange_name)
            return
        self.batch_ready.send(
            self,
            exchange_name=exchange_name,
            batch=batch,
        )
        del self.batches[exchange_name]

    def start_timer(self, exchange_name):
        """Start timer thread.

        A timer thread is started to make sure that the batch will be inserted
        into the database if the time limit is exceeded before the size limit.

        :param exchange_name: Exchange from which message batch was received
        :type exchange_name: str

        """
        if exchange_name in self.timers:
            LOGGER.warning('Timer already active for: %r', exchange_name)
            return
        timer = threading.Timer(
            self.time_limit,
            self.time_expired_cb,
            (exchange_name, ),
        )
        timer.name = 'timer-{}'.format(exchange_name)
        timer.daemon = True
        timer.start()
        LOGGER.debug(
            'Timer thread started (%.2f) for %r: (%d, %s)',
            self.time_limit,
            exchange_name,
            timer.ident,
            timer.name,
        )
        self.timers[exchange_name] = timer

    def cancel_timer(self, exchange_name):
        """Cancel timer thread.

        A timer thread might be cancelled if the size limit for a batch is
        exceeded before the time limit.

        :param exchange_name: Exchange from which message batch was received
        :type exchange_name: str

        """
        timer = self.timers.get(exchange_name)
        if timer is None:
            LOGGER.warning('Timer not found for: %r', exchange_name)
            return
        timer.cancel()
        LOGGER.debug(
            'Timer thread cancelled: (%d, %s)',
            timer.ident,
            timer.name,
        )
        del self.timers[exchange_name]



# -*- coding: utf-8 -*-

"""Consumer: get messages from rabbitmq.

The strategy to get messages is:
    - connect to the rabbitmq server
    - bind a queue to the desired exchanges

Note that it's assumed that the exchanges will have `fanout` type and that the
routing key isn't relevant in this case.

"""

import json
import logging

import blinker
import pika

LOGGER = logging.getLogger(__name__)


class Consumer(object):

    """Rabbitmq message consumer.

    :param server: Rabbitmq server IP address
    :type server: str
    :param exchange_names: Exchange names to bind to
    :type exchange_names: list(str)

    """

    def __init__(self, server, exchange_names):
        """Configure exchanges and queue."""
        LOGGER.info('Connecting to %r...', server)
        parameters = pika.ConnectionParameters(server)
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()

        # Use a single queue to process messages from all exchanges
        result = channel.queue_declare(auto_delete=True)
        queue_name = result.method.queue
        LOGGER.debug('Declared queue %r', queue_name)

        for exchange_name in exchange_names:
            channel.exchange_declare(
                exchange=exchange_name,
                exchange_type='fanout',
            )
            channel.queue_bind(
                exchange=exchange_name,
                queue=queue_name,
            )
            LOGGER.debug(
                'Queue %r bound to exchange %r', queue_name, exchange_name)

        channel.basic_consume(self.message_received_cb, queue=queue_name)

        self.channel = channel
        self.message_received = blinker.Signal()

    def run(self):
        """Run ioloop and consume messages."""
        logging.info('Waiting for messages...')
        self.channel.start_consuming()

    def message_received_cb(self, channel, method_frame, header_frame, body):
        """Handle message received.

        :param channel: Connection channel with rabbitmq server
        :type channel: pika.channel.Channel
        :param method_frame: AMPQ method related data
        :type method_frame: pika.spec.Deliver
        :param body: Message body
        :type body: str

        """
        exchange_name = method_frame.exchange
        LOGGER.debug('Message received from %r: %s', exchange_name, body)

        # Only accept json messages
        if header_frame.content_type != 'application/json':
            LOGGER.warning(
                'Message discarded. Unexpected content type: %r',
                header_frame.content_type,
            )
            channel.basic_nack(method_frame.delivery_tag, requeue=False)
            return

        channel.basic_ack(delivery_tag=method_frame.delivery_tag)
        try:
            payload = json.loads(body)
        except ValueError:
            LOGGER.warning('Body decoding error: %r', body)
        else:
            self.message_received.send(
                self,
                exchange_name=exchange_name,
                payload=payload,
            )



# -*- coding: utf-8 -*-

"""Batcher test cases."""

from unittest import TestCase

from mock import (
    MagicMock as Mock,
    patch,
)
from six.moves import range

from rabbithole.batcher import Batcher


class TestBatcher(TestCase):

    """Batcher test cases."""

    def setUp(self):
        """Create batcher object."""
        size_limit = 5
        time_limit = 15
        self.batcher = Batcher(size_limit, time_limit)

    def test_first_message_received(self):
        """Message appended to batch and timer started."""
        exchange = 'exchange'
        payload = 'payload'

        self.batcher.batch_ready = Mock()
        with patch('rabbithole.batcher.threading') as threading:
            self.batcher.message_received_cb('sender', exchange, payload)
            threading.Timer.assert_called_once_with(
                self.batcher.time_limit,
                self.batcher.time_expired_cb,
                (exchange, ),
            )
        self.assertListEqual(self.batcher.batches[exchange], [payload])

    def test_size_limit_exceeded(self):
        """Batch queued when size limit is exceed."""
        exchange = 'exchange'
        payload = 'payload'

        self.batcher.batch_ready = Mock()
        with patch('rabbithole.batcher.threading'):
            for _ in range(self.batcher.size_limit):
                self.batcher.message_received_cb('sender', exchange, payload)

        self.batcher.batch_ready.send.assert_called_with(
            self.batcher,
            exchange_name=exchange,
            batch=[payload] * self.batcher.size_limit,
        )
        self.assertListEqual(self.batcher.batches[exchange], [])

    def test_time_limit_exceeded(self):
        """Batch queued whem time limit is exceeded."""
        exchange = 'exchange'
        payload = 'payload'

        self.batcher.batch_ready = Mock()
        with patch('rabbithole.batcher.threading'):
            self.batcher.message_received_cb('sender', exchange, payload)
        self.batcher.time_expired_cb(exchange)

        self.batcher.batch_ready.send.assert_called_with(
            self.batcher,
            exchange_name=exchange,
            batch=[payload],
        )
        self.assertListEqual(self.batcher.batches[exchange], [])

    def test_expired_timer_not_found(self):
        """Warning written to logs when expired timer is not found."""
        exchange = 'exchange'

        with patch('rabbithole.batcher.LOGGER') as logger:
            self.batcher.time_expired_cb(exchange)
            logger.warning.assert_called_with(
                'Timer not found for: %r',
                exchange,
            )

    def test_timer_already_active(self):
        """Warning written to logs when timer is alreadya active."""
        exchange = 'exchange'

        self.batcher.timers[exchange] = Mock()
        with patch('rabbithole.batcher.LOGGER') as logger:
            self.batcher.start_timer(exchange)
            logger.warning.assert_called_with(
                'Timer already active for: %r',
                exchange,
            )

    def test_cancelled_timer_not_found(self):
        """Warning written to logs when cancelled timer is not found."""
        exchange = 'exchange'

        with patch('rabbithole.batcher.LOGGER') as logger:
            self.batcher.cancel_timer(exchange)
            logger.warning.assert_called_with(
                'Timer not found for: %r',
                exchange,
            )



# -*- coding: utf-8 -*-

"""CLI test cases."""

import argparse
import logging

from unittest import TestCase

import pika
import sqlalchemy
import yaml

from mock import patch
from six import StringIO

from rabbithole.cli import (
    configure_logging,
    main,
    parse_arguments,
)


class TestMain(TestCase):

    """Main entry point test cases."""

    def setUp(self):
        """Patch parse_arguments function."""
        parse_arguments_patcher = patch('rabbithole.cli.parse_arguments')
        parse_arguments_ = parse_arguments_patcher.start()
        parse_arguments_.return_value = argparse.Namespace(
            config={
                'rabbitmq': '<rabbitmq server>',
                'database': '<database url>',
                'output': {
                    'exchange#1': 'query#1',
                    'exchange#2': 'query#2',
                },
            },
            log_level=logging.CRITICAL,
        )
        self.addCleanup(parse_arguments_patcher.stop)

    def test_exit_on_rabbitmq_error(self):
        """Exit when there's a rabbitmq connectivity error."""
        with patch('rabbithole.cli.Consumer') as consumer_cls:
            consumer_cls.side_effect = pika.exceptions.AMQPError
            return_code = main()
            self.assertEqual(return_code, 1)

    def test_exit_on_database_error(self):
        """Exit when there's a database connectivity error."""
        with patch('rabbithole.cli.Consumer'), \
                patch('rabbithole.cli.Database') as database_cls:
            database_cls.side_effect = sqlalchemy.exc.SQLAlchemyError
            return_code = main()
            self.assertEqual(return_code, 1)

    def test_signals_connected(self):
        """Signals are connected as expected."""
        with patch('rabbithole.cli.Consumer') as consumer_cls, \
                patch('rabbithole.cli.Database') as database_cls, \
                patch('rabbithole.cli.Batcher') as batcher_cls:
            return_code = main()

            consumer_cls().message_received.connect.assert_called_once_with(
                batcher_cls().message_received_cb)
            batcher_cls().batch_ready.connect.assert_called_once_with(
                database_cls().connect().batch_ready_cb)

            self.assertEqual(return_code, 0)

    def test_exit_on_keyboard_interrupt(self):
        """Exit when user hits Ctrl+C."""
        with patch('rabbithole.cli.Consumer') as consumer_cls, \
                patch('rabbithole.cli.Database'), \
                patch('rabbithole.cli.Batcher'):
            consumer_cls().run.side_effect = KeyboardInterrupt
            return_code = main()
            self.assertEqual(return_code, 0)


class TestParseArguments(TestCase):

    """Argument parsing test cases."""

    def test_config_file_does_not_exist(self):
        """SystemExit is raised if the configuration file does not exist."""
        # Do not include error output in test output
        with self.assertRaises(SystemExit), patch('rabbithole.cli.sys.stderr'):
            parse_arguments(['file-does-not-exist'])

    def test_config_file_invalid(self):
        """SystemExit is raised if the configuration file is invalid."""
        with self.assertRaises(SystemExit), \
                patch('rabbithole.cli.sys.stderr'), \
                patch('rabbithole.cli.os') as os_, \
                patch('rabbithole.cli.open') as open_:
            os_.path.isfile.return_value = True
            open_().__enter__.return_value = StringIO('>invalid yaml<')
            parse_arguments(['some file'])

    def test_config_file_load_success(self):
        """Config file successfully loaded."""
        expected_value = {'a': 'value'}
        with patch('rabbithole.cli.os') as os_, \
                patch('rabbithole.cli.open') as open_:
            os_.path.isfile.return_value = True
            open_().__enter__.return_value = (
                StringIO(yaml.dump(expected_value)))
            args = parse_arguments(['some file'])

        self.assertDictEqual(args.config, expected_value)


class TestConfigureLogging(TestCase):

    """Logging configuration test cases."""

    def tearDown(self):
        """Delete root logger handlers."""
        logging.getLogger().handlers = []

    def test_root_level_set_to_debug(self):
        """Root logger level set to debug."""
        configure_logging(logging.ERROR)
        root_logger = logging.getLogger()
        self.assertEqual(root_logger.level, logging.DEBUG)

    def test_stream_handler_level(self):
        """Stream handler level set to argument value."""
        expected_value = logging.ERROR
        configure_logging(expected_value)
        root_logger = logging.getLogger()
        self.assertEqual(len(root_logger.handlers), 1)
        handler = root_logger.handlers[0]
        self.assertEqual(handler.level, expected_value)

    def test_pika_level_set_warning(self):
        """Pika logger level is set to warning."""
        configure_logging(logging.DEBUG)
        pika_logger = logging.getLogger('pika')
        self.assertEqual(pika_logger.level, logging.WARNING)



# -*- coding: utf-8 -*-

"""Store messages from rabbitmq into a SQL database."""

import argparse
import logging
import os
import sys

from pprint import pformat

import pika
import sqlalchemy
import yaml

from rabbithole.consumer import Consumer
from rabbithole.db import Database
from rabbithole.batcher import Batcher

LOGGER = logging.getLogger(__name__)


def main(argv=None):
    """Console script for rabbithole

    :param argv: Command line arguments
    :type argv: list(str)

    """
    if argv is None:
        argv = sys.argv[1:]

    args = parse_arguments(argv)
    config = args.config
    configure_logging(args.log_level)
    logging.debug('Configuration:\n%s', pformat(config))

    try:
        consumer = Consumer(config['rabbitmq'], config['output'].keys())
    except pika.exceptions.AMQPError as exception:
        LOGGER.error('Rabbitmq connectivity error: %s', exception)
        return 1

    try:
        database = Database(config['database'], config['output']).connect()
    except sqlalchemy.exc.SQLAlchemyError as exception:
        LOGGER.error(exception)
        return 1

    batcher = Batcher(config.get('size_limit'), config.get('time_limit'))
    consumer.message_received.connect(batcher.message_received_cb)
    batcher.batch_ready.connect(database.batch_ready_cb)

    try:
        consumer.run()
    except KeyboardInterrupt:
        LOGGER.info('Interrupted by user')

    return 0


def parse_arguments(argv):
    """Parse command line arguments.

    :returns: Parsed arguments
    :rtype: argparse.Namespace

    """
    parser = argparse.ArgumentParser(description=__doc__)

    def yaml_file(path):
        """Yaml file argument.

        :param path: Path to the yaml file
        :type path: str

        """
        if not os.path.isfile(path):
            raise argparse.ArgumentTypeError('File not found')

        with open(path) as file_:
            try:
                data = yaml.load(file_)
            except yaml.YAMLError:
                raise argparse.ArgumentTypeError('YAML parsing error')

        return data

    parser.add_argument(
        'config',
        type=yaml_file,
        help='Configuration file',
    )

    log_levels = ['debug', 'info', 'warning', 'error', 'critical']
    parser.add_argument(
        '-l', '--log-level',
        dest='log_level',
        choices=log_levels,
        default='debug',
        help=('Log level. One of {0} or {1} '
              '(%(default)s by default)'
              .format(', '.join(log_levels[:-1]), log_levels[-1])))

    args = parser.parse_args(argv)
    args.log_level = getattr(logging, args.log_level.upper())
    return args


def configure_logging(log_level):
    """Configure logging based on command line argument.

    :param log_level: Log level passed form the command line
    :type log_level: int

    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    # Log to sys.stderr using log level
    # passed through command line
    log_handler = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s %(threadName)s %(levelname)s: %(message)s')
    log_handler.setFormatter(formatter)
    log_handler.setLevel(log_level)
    root_logger.addHandler(log_handler)

    # Disable pika extra verbose logging
    logging.getLogger('pika').setLevel(logging.WARNING)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""Database: run queries with batches of rows per exchange."""

import logging

from sqlalchemy import (
    create_engine,
    text,
)
from sqlalchemy.exc import SQLAlchemyError

LOGGER = logging.getLogger(__name__)


class Database(object):

    """Database writer.

    :param url: Database connection string
    :type url: str
    :param queries: Insert query to execute for each exchange name
    :type queries: dict(str, str)

    """

    def __init__(self, url, queries):
        """Connect to database."""
        self.engine = create_engine(url)
        self.connection = None

        self.queries = {
            exchange_name: text(query)
            for exchange_name, query
            in queries.items()
        }

    def connect(self):
        """Connect to the database."""
        self.connection = self.engine.connect()
        LOGGER.debug('Connected to: %r', self.engine.url)
        return self

    def batch_ready_cb(self, sender, exchange_name, batch):
        """Execute insert query for the batch that is ready.

        :param sender: The batcher who sent the batch_ready signal
        :type sender: rabbithole.batcher.Batcher
        :param exchange_name: Key used to determine which query to execute
        :type exchange_name: str
        :param batch: Batch of rows to insert
        :type rows: list(dict(str))

        """
        if exchange_name not in self.queries:
            LOGGER.warning('No query found for %r', exchange_name)
            return

        query = self.queries[exchange_name]
        try:
            self.connection.execute(query, batch)
        except SQLAlchemyError as exception:
            LOGGER.error(exception)
        else:
            LOGGER.debug('Inserted %d rows', len(batch))



# -*- coding: utf-8 -*-

"""Database output test cases."""

from unittest import TestCase

from mock import (
    MagicMock as Mock,
    patch,
)
from sqlalchemy.exc import SQLAlchemyError

from rabbithole.db import Database


class TestConnect(TestCase):

    """Database connection test cases."""

    def test_is_connected(self):
        """Connection is open after call to connect."""
        database = Database('sqlite://', {}).connect()
        self.assertFalse(database.connection.closed)


class TestBatchReady(TestCase):

    """Database batch ready test cases."""

    def setUp(self):
        """Create database object."""
        self.database = Database(
            'sqlite://',
            {
                'exchange#1': 'query#1',
                'exchange#2': 'query#2',
            },
        ).connect()

    def test_warning_on_query_not_found(self):
        """Warning written to logs if no query found."""
        exchange_name = '<exchange>'

        with patch('rabbithole.db.LOGGER') as logger:
            self.database.batch_ready_cb('<sender>', exchange_name, [])
            logger.warning.assert_called_once_with(
                'No query found for %r',
                exchange_name,
            )

    def test_query_executed(self):
        """Query executed for the given exchange."""
        exchange_name = 'exchange#1'
        batch = [1, 2, 3]

        self.database.connection = Mock()
        self.database.batch_ready_cb('<sender>', exchange_name, batch)
        self.database.connection.execute.assert_called_once_with(
            self.database.queries[exchange_name],
            batch,
        )

    def test_eror_on_database_error(self):
        """Error written to logs on query execution failure."""
        exchange_name = 'exchange#1'
        batch = [1, 2, 3]
        exception = SQLAlchemyError('<error>')

        self.database.connection = Mock()
        self.database.connection.execute.side_effect = exception
        with patch('rabbithole.db.LOGGER') as logger:
            self.database.batch_ready_cb('<sender>', exchange_name, batch)
            logger.error.assert_called_once_with(exception)



# -*- coding: utf-8 -*-

"""Message consumer test cases."""

import json

from unittest import TestCase

from mock import (
    MagicMock as Mock,
    patch,
)

from rabbithole.consumer import Consumer


class TestConsumer(TestCase):

    """Consumer test cases."""

    QUEUE = '<queue_name>'
    EXCHANGES = [
        'exchange#1',
        'exchange#2',
    ]

    def setUp(self):
        """Create fake channel."""
        pika_patcher = patch('rabbithole.consumer.pika')
        pika = pika_patcher.start()
        self.addCleanup(pika_patcher.stop)

        self.channel = pika.BlockingConnection().channel()

    def test_exchanges_declared(self):
        """Consumer declares exhanges on initialization"""
        Consumer('<server>', self.EXCHANGES)
        for exchange in self.EXCHANGES:
            self.channel.exchange_declare.assert_any_call(
                exchange=exchange,
                exchange_type='fanout',
            )

    def test_queue_bound(self):
        """Queue is bound to the exchanges."""
        self.channel.queue_declare().method.queue = self.QUEUE
        Consumer('<server>', self.EXCHANGES)
        for exchange in self.EXCHANGES:
            self.channel.queue_bind.assert_any_call(
                exchange=exchange,
                queue=self.QUEUE,
            )

    def test_run(self):
        """Consumer starts consuming when run is called."""
        consumer = Consumer('<server>', self.EXCHANGES)
        consumer.run()
        self.channel.start_consuming.assert_called_once_with()

    def test_message_content_type(self):
        """Message is discarded based on content_type."""
        consumer = Consumer('<server>', self.EXCHANGES)

        channel = Mock()
        method_frame = Mock()
        header_frame = Mock()
        header_frame.content_type = 'text/plain'
        consumer.message_received_cb(
            channel,
            method_frame,
            header_frame,
            '<body>',
        )

        channel.basic_nack.assert_called_with(
            method_frame.delivery_tag,
            requeue=False,
        )

    def test_message_decoding_error(self):
        """Warning written to logs when body cannot be decoded."""
        body = '<body>'
        consumer = Consumer('<server>', self.EXCHANGES)

        consumer.message_received = Mock()
        channel = Mock()
        method_frame = Mock()
        header_frame = Mock()
        header_frame.content_type = 'application/json'

        with patch('rabbithole.consumer.LOGGER') as logger:
            consumer.message_received_cb(
                channel,
                method_frame,
                header_frame,
                body,
            )
            logger.warning.assert_called_once_with(
                'Body decoding error: %r',
                body,
            )

    def test_message_received(self):
        """Message received signal is sent if message is correct."""
        body = '<body>'
        consumer = Consumer('<server>', self.EXCHANGES)

        consumer.message_received = Mock()
        channel = Mock()
        method_frame = Mock()
        header_frame = Mock()
        header_frame.content_type = 'application/json'
        consumer.message_received_cb(
            channel,
            method_frame,
            header_frame,
            json.dumps(body),
        )

        channel.basic_ack.assert_called_with(
            delivery_tag=method_frame.delivery_tag)
        consumer.message_received.send.assert_called_with(
            consumer,
            exchange_name=method_frame.exchange,
            payload=body,
        )
